
import bdb
import re
import pdb
import sys
import linecache
from hrepr import HTML
from .repr import hrepr


H = HTML()


class BucheDb(bdb.Bdb):
    # Note: Many methods have been lifted from pdb.py in the standard lib
    # and slightly adapted

    def __init__(self, repl):
        super().__init__()
        self.repl = repl
        self.frame = None
        self.display_frame = True
        self.stack = []
        self.current = 0

    def eval(self, repl, code, code_globals):
        frame = self.get_frame()
        gs = {**frame.f_globals, **code_globals.globals}
        ls = frame.f_locals
        m = re.match(r'^[?!]+', code)
        if m:
            cmd = code[:m.end()]
            arg = code[m.end():]
        elif ' ' in code:
            cmd, arg = code.split(' ', 1)
        else:
            cmd = code
            arg = ''
        method = self.__commands__.get(cmd, None)
        if method:
            return method(self, arg, gs, ls)
        else:
            return self.command_print(code, gs, ls)

    def setup(self, frame, tb):
        self.stack, self.current = self.get_stack(frame, tb)

    def checkline(self, filename, lineno):
        """Check whether specified line seems to be executable.

        Return `lineno` if it is, 0 if not (e.g. a docstring, comment, blank
        line or EOF). Warning: testing is not comprehensive.
        """
        # this method should be callable before starting debugging, so default
        # to "no globals" if there is no current frame
        curframe = self.get_frame()
        globs = curframe.f_globals if curframe else None
        line = linecache.getline(filename, lineno, globs)
        if not line:
            self.message('End of file')
            return 0
        line = line.strip()
        # Don't allow setting breakpoint at a blank line
        if (not line or (line[0] == '#') or
             (line[:3] == '"""') or line[:3] == "'''"):
            self.error('Blank or comment')
            return 0
        return lineno

    def print_stack_trace(self):
        res = H.boxTabs()
        stack = self.stack[:self.current + 1]
        for i, (entry, lineno) in enumerate(stack):
            code = entry.f_code
            res = res(H.tabEntry(
                hrepr(entry),
                label=f'{code.co_name}@{code.co_filename}:{entry.f_lineno}',
                active=(i == len(stack) - 1)
            ))
        self.repl.log(res, interactive=True)

    def command_step(self, arg, gs, ls):
        """<b>s(tep)</b>

        Execute the current line, stop at the first possible occasion
        (either in a function that is called or in the current function).
        """
        self.set_step()
        self.proceed = True

    def command_next(self, arg, gs, ls):
        """<b>n(ext)</b>

        Continue execution until the next line in the current function
        is reached or it returns.
        """
        self.set_next(self.get_frame())
        self.proceed = True

    def command_continue(self, arg, gs, ls):
        """<b>c(ont(inue))</b>

        Continue execution, only stop when a breakpoint is encountered.
        """
        self.set_continue()
        self.proceed = True

    def command_return(self, arg, gs, ls):
        """<b>r(eturn)</b>

        Continue execution until the current function returns.
        """
        self.set_return()
        self.proceed = True

    def command_until(self, arg, gs, ls):
        """<b>unt(il) [lineno]</b>

        Without argument, continue execution until the line with a
        number greater than the current one is reached.  With a line
        number, continue execution until a line with a number greater
        or equal to that is reached.  In both cases, also stop when
        the current frame returns.
        """
        curframe = self.get_frame()
        if arg:
            try:
                lineno = int(arg)
            except ValueError:
                self.error('Error in argument: %r' % arg)
                return
            if lineno <= curframe.f_lineno:
                self.error('"until" line number is smaller than current '
                           'line number')
                return
        else:
            lineno = None
        self.set_until(curframe, lineno)
        self.proceed = True

    def command_up(self, arg, gs, ls):
        """<b>u(p) [count]</b>

        Move the current frame count (default one) levels up in the
        stack trace (to an older frame).
        """
        if self.current == 0:
            self.error('Oldest frame')
            return
        try:
            count = int(arg or 1)
        except ValueError:
            self.error('Invalid frame count (%s)' % arg)
            return
        if count < 0:
            self.current = 0
        else:
            self.current = max(0, self.current - count)
        self.repl.log(self.get_frame())

    def command_down(self, arg, gs, ls):
        """<b>d(own) [count]</b>

        Move the current frame count (default one) levels down in the
        stack trace (to a newer frame).
        """
        if self.current + 1 == len(self.stack):
            self.error('Newest frame')
            return
        try:
            count = int(arg or 1)
        except ValueError:
            self.error('Invalid frame count (%s)' % arg)
            return
        if count < 0:
            self.current = len(self.stack) - 1
        else:
            self.current = min(len(self.stack) - 1, self.current + count)
        self.repl.log(self.get_frame())

    def command_break(self, arg, gs, ls, temporary = 0):
        """<b>b(reak) [ ([filename:]lineno | function) [, condition] ]</b>

        Without argument, list all breaks.

        With a line number argument, set a break at this line in the
        current file.  With a function name, set a break at the first
        executable line of that function.  If a second argument is
        present, it is a string specifying an expression which must
        evaluate to true before the breakpoint is honored.

        The line number may be prefixed with a filename and a colon,
        to specify a breakpoint in another file (probably one that
        hasn't been loaded yet).  The file is searched for on
        sys.path; the .py suffix may be omitted.
        """
        curframe = self.get_frame()
        if not arg:
            if self.breaks:
                t = H.table()
                t = t(H.tr(
                    H.th('Num'),
                    H.th('Disp'),
                    H.th('Enb'),
                    H.th('Where'),
                ))
                for bp in bdb.Breakpoint.bpbynumber:
                    if bp:
                        disp = 'del' if bp.temporary else 'keep'
                        enb = 'yes' if bp.enabled else 'no'
                        t = t(H.tr(
                            H.td(bp.number),
                            H.td(disp),
                            H.td(enb),
                            H.td(f'{bp.file}:{bp.line}'),
                        ))
                self.repl.log.html(t)
            return
        # parse arguments; comma has lowest precedence
        # and cannot occur in filename
        filename = None
        lineno = None
        cond = None
        comma = arg.find(',')
        if comma > 0:
            # parse stuff after comma: "condition"
            cond = arg[comma+1:].lstrip()
            arg = arg[:comma].rstrip()
        # parse stuff before comma: [filename:]lineno | function
        colon = arg.rfind(':')
        funcname = None
        if colon >= 0:
            filename = arg[:colon].rstrip()
            f = self.lookupmodule(filename)
            if not f:
                self.error('%r not found from sys.path' % filename)
                return
            else:
                filename = f
            arg = arg[colon+1:].lstrip()
            try:
                lineno = int(arg)
            except ValueError:
                self.error('Bad lineno: %s' % arg)
                return
        else:
            # no colon; can be lineno or function
            try:
                lineno = int(arg)
            except ValueError:
                try:
                    func = eval(arg,
                                curframe.f_globals,
                                curframe_locals)
                except:
                    func = arg
                try:
                    if hasattr(func, '__func__'):
                        func = func.__func__
                    code = func.__code__
                    #use co_name to identify the bkpt (function names
                    #could be aliased, but co_name is invariant)
                    funcname = code.co_name
                    lineno = code.co_firstlineno
                    filename = code.co_filename
                except:
                    # last thing to try
                    (ok, filename, ln) = self.lineinfo(arg)
                    if not ok:
                        self.error('The specified object %r is not a function '
                                   'or was not found along sys.path.' % arg)
                        return
                    funcname = ok # ok contains a function name
                    lineno = int(ln)
        if not filename:
            filename = curframe.f_code.co_filename
        # Check for reasonable breakpoint
        line = self.checkline(filename, lineno)
        if line:
            # now set the break point
            err = self.set_break(filename, line, temporary, cond, funcname)
            if err:
                self.error(err)
            else:
                bp = self.get_breaks(filename, line)[-1]
                self.message("Breakpoint %d at %s:%d" %
                             (bp.number, bp.file, bp.line))

    def command_tbreak(self, arg, gs, ls):
        """<b>tbreak [ ([filename:]lineno | function) [, condition] ]</b>

        Same arguments as break, but sets a temporary breakpoint: it
        is automatically deleted when first hit.
        """
        self.command_break(arg, gs, ls, 1)

    def command_enable(self, arg, gs, ls):
        """<b>enable bpnumber [bpnumber ...]</b>

        Enables the breakpoints given as a space separated list of
        breakpoint numbers.
        """
        args = arg.split()
        for i in args:
            try:
                bp = self.get_bpbynumber(i)
            except ValueError as err:
                self.error(err)
            else:
                bp.enable()
                self.message('Enabled %s' % bp)

    def command_disable(self, arg, gs, ls):
        """<b>disable bpnumber [bpnumber ...]</b>

        Disables the breakpoints given as a space separated list of
        breakpoint numbers.  Disabling a breakpoint means it cannot
        cause the program to stop execution, but unlike clearing a
        breakpoint, it remains in the list of breakpoints and can be
        (re-)enabled.
        """
        args = arg.split()
        for i in args:
            try:
                bp = self.get_bpbynumber(i)
            except ValueError as err:
                self.error(err)
            else:
                bp.disable()
                self.message('Disabled %s' % bp)

    def command_condition(self, arg, gs, ls):
        """<b>condition bpnumber [condition]</b>

        Set a new condition for the breakpoint, an expression which
        must evaluate to true before the breakpoint is honored.  If
        condition is absent, any existing condition is removed; i.e.,
        the breakpoint is made unconditional.
        """
        args = arg.split(' ', 1)
        try:
            cond = args[1]
        except IndexError:
            cond = None
        try:
            bp = self.get_bpbynumber(args[0].strip())
        except IndexError:
            self.error('Breakpoint number expected')
        except ValueError as err:
            self.error(err)
        else:
            bp.cond = cond
            if not cond:
                self.message('Breakpoint %d is now unconditional.' % bp.number)
            else:
                self.message('New condition set for breakpoint %d.' % bp.number)

    def command_ignore(self, arg, gs, ls):
        """<b>ignore bpnumber [count]</b>

        Set the ignore count for the given breakpoint number.  If
        count is omitted, the ignore count is set to 0.  A breakpoint
        becomes active when the ignore count is zero.  When non-zero,
        the count is decremented each time the breakpoint is reached
        and the breakpoint is not disabled and any associated
        condition evaluates to true.
        """
        args = arg.split()
        try:
            count = int(args[1].strip())
        except:
            count = 0
        try:
            bp = self.get_bpbynumber(args[0].strip())
        except IndexError:
            self.error('Breakpoint number expected')
        except ValueError as err:
            self.error(err)
        else:
            bp.ignore = count
            if count > 0:
                if count > 1:
                    countstr = '%d crossings' % count
                else:
                    countstr = '1 crossing'
                self.message('Will ignore next %s of breakpoint %d.' %
                             (countstr, bp.number))
            else:
                self.message('Will stop next time breakpoint %d is reached.'
                             % bp.number)

    def command_clear(self, arg, gs, ls):
        """<b>cl(ear) filename:lineno\ncl(ear) [bpnumber [bpnumber...]]</b>
    
        With a space separated list of breakpoint numbers, clear
        those breakpoints.  Without argument, clear all breaks (but
        first ask confirmation).  With a filename:lineno argument,
        clear all breaks at that line in that file.
        """
        if not arg:
            try:
                reply = input('Clear all breaks? ')
            except EOFError:
                reply = 'no'
            reply = reply.strip().lower()
            if reply in ('y', 'yes'):
                bplist = [bp for bp in bdb.Breakpoint.bpbynumber if bp]
                self.clear_all_breaks()
                for bp in bplist:
                    self.message('Deleted %s' % bp)
            return
        if ':' in arg:
            # Make sure it works for "clear C:\foo\bar.py:12"
            i = arg.rfind(':')
            filename = arg[:i]
            arg = arg[i+1:]
            try:
                lineno = int(arg)
            except ValueError:
                err = "Invalid line number (%s)" % arg
            else:
                bplist = self.get_breaks(filename, lineno)
                err = self.clear_break(filename, lineno)
            if err:
                self.error(err)
            else:
                for bp in bplist:
                    self.message('Deleted %s' % bp)
            return
        numberlist = arg.split()
        for i in numberlist:
            try:
                bp = self.get_bpbynumber(i)
            except ValueError as err:
                self.error(err)
            else:
                self.clear_bpbynumber(i)
                self.message('Deleted %s' % bp)

    def command_where(self, arg, gs, ls):
        """<b>w(here)</b>

        Print a stack trace, with the most recent frame at the bottom.
        An arrow indicates the "current frame", which determines the
        context of most commands.  'bt' is an alias for this command.
        """
        self.print_stack_trace()

    def command_print(self, code, gs, ls):
        """<b>p(rint) expression</b>

        Print the value of the expression. Synonyms: !, pp
        """
        try:
            return eval(code, gs, ls)
        except SyntaxError:
            return exec(code, gs, ls)

    def command_quit(self, arg, gs, ls):
        """<b>q(uit)</b>

        Quit from the debugger. The program being executed is aborted.
        """
        self._user_requested_quit = True
        self.set_quit()
        self.proceed = True

    def command_help(self, arg, gs, ls):
        """<b>h(elp)</b>

        Without argument, print the list of available commands.
        With a command name as argument, print help about that command.
        """
        if arg:
            self.repl.log.html(self.__commands__[arg].__doc__)
        else:
            commands = [x.__doc__ for x in set(self.__commands__.values())]
            commands.sort()
            for cmd in commands:
                self.repl.log.html(cmd)

    __commands__ = {
        'step': command_step,
        's': command_step,
        'next': command_next,
        'n': command_next,
        'continue': command_continue,
        'c': command_continue,
        'return': command_return,
        'r': command_return,
        'until': command_until,
        'unt': command_until,
        'up': command_up,
        'u': command_up,
        'down': command_down,
        'd': command_down,
        'print': command_print,
        'p': command_print,
        'pp': command_print,
        '!': command_print,
        'help': command_help,
        'h': command_help,
        '?': command_help,
        'break': command_break,
        'tbreak': command_tbreak,
        'enable': command_enable,
        'disable': command_disable,
        'condition': command_condition,
        'ignore': command_ignore,
        'clear': command_clear,
        'cl': command_clear,
        'where': command_where,
        'w': command_where,
        'bt': command_where,
        'quit': command_quit,
        'q': command_quit,
    }

    def error(self, message):
        self.repl.log.html.logEntry['error'](message)

    def message(self, message):
        self.repl.log.html.logEntry(message)

    def get_frame(self):
        return self.stack[self.current][0]

    def set_frame(self, frame, tb=None):
        self.setup(frame, tb)
        self.repl.log(self.get_frame())
        self.proceed = False
        while not self.proceed:
            self.repl.query(eval=self.eval)

    # def user_call(self, frame, args):
    #     self.repl.log.html('<b>Enter call</b>')
    #     self.set_frame(frame)

    def user_line(self, frame):
        self.repl.log.html('<b>Next line</b>')
        self.set_frame(frame)

    # def user_return(self, frame, rval):
    #     self.repl.log.html('<b>Return</b>')
    #     self.set_frame(frame)

    def user_exception(self, frame, exc_info):
        self.repl.log.html('<b>An exception occurred</b>')
        self.repl.log(exc_info)
        self.set_frame(frame)

    def set_trace(self, frame=None):
        self.repl.start(synchronous=True)
        super().set_trace(frame or sys._getframe(1))

    def interaction(self, frame, tb):
        self.repl.start(synchronous=True)
        self.set_frame(frame, tb)




import os
import traceback
from hrepr import StdHRepr, Tag
from weakref import WeakValueDictionary
from collections import deque
from types import FrameType


class UNAVAILABLE:
    def __init__(self):
        pass

    def __hrepr__(self, H, hrepr):
        return H.span('<UNAVAILABLE>', style='color:#bbb')


UNAVAILABLE = UNAVAILABLE()


class Registry:
    def __init__(self, keep=10000):
        self.keep = keep
        self.id = 0
        self.weak_map = WeakValueDictionary()
        self.strong_map = {}
        self.strong_ids = deque()

    def register(self, obj):
        self.id += 1
        try:
            self.weak_map[self.id] = obj
        except TypeError:
            self.strong_map[self.id] = obj
            self.strong_ids.append(self.id)
            if len(self.strong_ids) > self.keep >= 0:
                rm = self.strong_ids.popleft()
                del self.strong_map[rm]
        return self.id

    def resolve(self, id):
        try:
            return self.weak_map[id]
        except KeyError:
            return self.strong_map.get(id, UNAVAILABLE)


id_registry = Registry()


def handle_exception(e, H, hrepr):
    tb = e.__traceback__
    entries = traceback.extract_tb(tb)
    first = e.args[0] if len(e.args) > 0 else None
    iss = isinstance(first, str)
    args_table = H.table()
    for i in range(1 if iss else 0, len(e.args)):
        arg = e.args[i]
        tr = H.tr(H.td(i), H.td(hrepr(arg)))
        args_table = args_table(tr)

    views = H.boxTabs()

    if entries:
        last = entries[-1]
        for entry in entries:
            filename, lineno, funcname, line = entry
            absfile = os.path.abspath(filename)
            view = H.tabEntry()
            tab = H.tabLabel(
                f'{funcname}@{os.path.basename(absfile)}'
            )
            snippet = H.codeSnippet(
                src = absfile,
                language = "python",
                line = lineno,
                context = 4
            )
            if filename.startswith('<'):
                snippet = snippet(getattr(e, '__repl_string__', ""))

            pane = H.tabPane(snippet)
            if entry is last:
                view = view(active = True)
            views = views(view(tab, pane))

    container = H.div['hrepr-Exception'](
        H.div(
            H.strong(type(e).__name__),
            f': {first}' if iss else '',
            args_table
        ),
        views
    )

    return container


def handle_exception_resources(H):
    return H.style('''
        .hrepr-Exception {
            display: inline-block;
            border: 2px solid red;
        }
        .hrepr-Exception box-tabs {
            border: 0px;
            border-top: 1px solid black;
        }
    ''')


handle_exception.resources = handle_exception_resources


def handle_frame(fr, H, hrepr):
    views = H.boxTabs()
    code = fr.f_code
    info = dict(
        name=code.co_name,
        filename=code.co_filename,
        firstlineno=code.co_firstlineno,
        lineno=fr.f_lineno
    )
    snippet = H.codeSnippet(
        src = code.co_filename,
        language = "python",
        line = fr.f_lineno,
        context = 4
    )
    views = views(H.tabEntry(H.tabLabel('code'),
                             H.tabPane(snippet),
                             active=True))
    views = views(H.tabEntry(H.tabLabel('info'), H.tabPane(hrepr(info))))
    return views


class HRepr(StdHRepr):
    def __default_handlers__(self):
        h = super().__default_handlers__()
        h.update({
            Exception: handle_exception,
            FrameType: handle_frame
        })
        return h

    def hrepr_nowrap(self, obj, **kwargs):
        return super().__call__(obj, **kwargs)

    def __call__(self, obj, **kwargs):
        interactive = kwargs.get('interactive', False) \
            or self.config.interactive
        res = super().__call__(obj, **kwargs)
        if not interactive \
                or not isinstance(res, Tag) \
                or res.attributes.get('interactive', False):
            return res
        try:
            the_id = id_registry.register(obj)
        except TypeError as e:
            return res
        else:
            return res({'onclick': f"bucheSend(event, '{the_id}')"})


hrepr = HRepr()




import os
import sys

errmsg = """ERROR: The buche command is not installed.

The buche command is a language-agnostic application and has to be
installed separately from the Python package. You can find the latest
release for your OS here:

    https://github.com/breuleux/buche/releases/latest

Or if npm is installed, you can run the command:

    npm install buche -g
"""

if __name__ == '__main__':
    os.environ['PYTHONBREAKPOINT'] = 'buche.breakpoint'
    try:
        os.execvp('buche', ['buche', 'python3', '-u'] + sys.argv[1:])
    except FileNotFoundError:
        print(errmsg, file=sys.stderr)
        sys.exit(1)



"""
Simple Event and EventDispatcher.

These allow Myia's various components and transformations to
inform watchers about what they are doing. These watchers can
then identify problems, log actions, etc.
"""

from typing import Any, Union, Callable


Handler = Union['Event', Callable]


class Event(list):
    """
    Simple Event class.

    >>> e = Event('tickle')
    >>> e.register(lambda event, n: print("hi" * n))
    <function>
    >>> e(5)
    hihihihihi

    Event's constructor takes an additional argument, which
    is the EventDispatcher to which it belongs (or None).
    """
    def __init__(self,
                 name: str,
                 dispatcher: 'EventDispatcher' = None) -> None:
        self.name = name
        self.dispatcher = dispatcher
        self.owner: Any = dispatcher.owner if dispatcher else None

    def register(self, handler: Handler) -> Handler:
        # Returns the handler so it can be used as a decorator
        self.append(handler)
        return handler

    def __call__(self, *args, **kwargs) -> Any:
        if len(args) > 0 and not isinstance(args[0], Event):
            args = (self,) + args
        for f in self:
            f(*args, **kwargs)

    def __str__(self):
        return f'Event({self.name})'

    def __repr__(self):
        return str(self)

    def __hrepr__(self):
        return H.span['Event'](f'Event({self.name})')


class EventDispatcher:
    """
    Handle arbitrary named events.

    * ``on_<event_name>(...)`` is shorthand for
      ``on(event_name, ...)``. Tip: it can be used as a
      decorator.
    * ``emit_<event_name>(...)`` is shorthand for
      ``emit(event_name, ...)``

    >>> d = EventDispatcher()
    >>> d.on_tickle(lambda event, n: print("hi" * n))
    <function>
    >>> d.emit_tickle(6)
    hihihihihihi

    An EventDispatcher also provides two special events:

    * ``NEW`` (all caps) is emitted whenever a new event
      is created (the first time ``on(event_name, ...)``
      is called for a given event_name).
    * ``ALL`` is triggered for all events.

    Tip: if you want to make event A fire when event B fires,
    you can write: ``d.on_B(d.emit_A)``.
    """

    def __init__(self, owner: Any = None, discover=True) -> None:
        self.owner = owner
        self._events = {'NEW': Event('NEW', self),
                        'ALL': Event('ALL', self)}
        if discover:
            discovery_event(owner or self, self)

    def on(self, event_name: str, *handlers: Callable) -> None:
        """
        Register one or more handlers
        """
        self[event_name].extend(handlers)

    def emit(self, event_name: str, *args, **kwargs) -> None:
        self[event_name](*args, **kwargs)

    def __getitem__(self, event_name: str) -> 'Event':
        e = self._events.get(event_name, None)
        if e is None:
            e = Event(event_name, self)
            self._events[event_name] = e
            self._events['NEW'](event_name, e)
            e.register(self._events['ALL'])
        return e

    def __getattr__(self, event_name: str) -> Union[Callable, Event]:
        if event_name.startswith('on_'):
            return self[event_name[3:]].register
        elif event_name.startswith('emit_'):
            return self[event_name[5:]]
        else:
            raise AttributeError(event_name)


# Newly constructed EventDispatchers emit this event so that
# we can hook to them more easily.
# Event signature: (discovery_event, dispatcher_owner, dispatcher)
discovery_event = Event('discovery')


def on_discovery(type: Any, event_name: str = None) -> Callable:
    """
    Mechanism to "discover" events. Given a type and an event name,
    ``on_discovery`` makes sure that your function triggers when
    that event is emitted by any instance of the given type created
    *after* the ``on_discovery`` call.

    The event name can be given explicitly, but if you decorate a
    function called ``on_<event_name>``, ``on_discovery`` will
    infer that this is the event you want to listen to.

    For example:

    >>> class MyClass(EventDispatcher): pass
    >>> @on_discovery(MyClass)
    ... def on_dog(*args):
    ...     print('I like doggies!')
    >>> c = MyClass()
    >>> c.emit_dog()
    I like doggies!
    """
    def decorate(fn: Callable) -> Callable:
        nonlocal event_name
        if event_name is None and fn.__name__.startswith('on_'):
            event_name = fn.__name__[3:]
        assert event_name is not None

        def seek(_, owner, dispatcher):
            if isinstance(owner, type):
                dispatcher.on(event_name, fn)
        discovery_event.register(seek)
        return fn
    return decorate




from uuid import uuid4


class CodeGlobals:
    def __init__(self, glb=None):
        self.globals = glb or {}
        self.saved = []

    def getvar(self, obj):
        try:
            rank = self.saved.index(obj)
            varname = f'_{rank + 1}'
        except ValueError:
            self.saved.append(obj)
            rank = len(self.saved)
            varname = f'_{rank}'
            self.globals[varname] = obj
        return varname


class Repl:

    @staticmethod
    def default_evaluate(repl, code, code_globals):
        glob = code_globals.globals
        try:
            return eval(code, glob)
        except SyntaxError:
            exec(code, glob)

    def __init__(self,
                 buche,
                 reader,
                 evaluator=None,
                 code_globals=None,
                 address=None,
                 log_address=None):
        self.buche = buche
        self.reader = reader
        self.code_globals = code_globals or CodeGlobals()
        if not isinstance(self.code_globals, CodeGlobals):
            self.code_globals = CodeGlobals(self.code_globals)
        self.address = address or str(uuid4())
        self.relative_log_address = log_address or 'log'
        self.log_address = log_address or f'{self.address}/log'
        self.cmdidx = 0
        self.cmdlog = ['']
        self.input_address = f'{self.address}/input'
        self.input = buche[self.address]['input']
        self.log = buche[self.log_address]
        self.code_globals.globals['buche'] = self.log
        self.eval = evaluator or self.default_evaluate
        self.started = False

    def _repl_key(self, event, message):
        if not message.path.endswith(self.input_address):
            return

        key = message.which
        if key == 'Up':
            self.cmdidx = (self.cmdidx - 1) % len(self.cmdlog)
            self.input.command_set(value=self.cmdlog[self.cmdidx])
        elif key == 'Down':
            self.cmdidx = (self.cmdidx + 1) % len(self.cmdlog)
            self.input.command_set(value=self.cmdlog[self.cmdidx])

    def _repl_line(self, event, message):
        return self.handle_message(message)

    def _repl_select(self, event, message):
        if self.log_address not in message.enclosingPath:
            return

        if hasattr(message, 'obj'):
            varname = self.code_globals.getvar(message.obj)
            self.input.command_append(value=varname)
            self.input.command_focus()

    def handle_message(self, message, eval=None):
        if message.eventType != 'submit':
            return False
        if not message.path.endswith(self.input_address):
            return False
        return self.run(message.value, eval)

    def run(self, code, eval=None):
        if code.strip() == '':
            return False
        self.log.html.logEntry['echo'](code)
        self.input.command_set()
        try:
            res = (eval or self.eval)(self, code, self.code_globals)
        except Exception as exc:
            self.log.show.logEntry['error'](exc)
        else:
            if res is not None:
                self.log.show.logEntry['result'](res, interactive=True)
        self.cmdlog.append(code)
        self.cmdidx = 0
        return True

    def start(self, nodisplay=False, synchronous=False):
        if self.started:
            return

        if not nodisplay:
            self.buche(self)

        if not synchronous:
            self.reader.on_submit(self._repl_line)
        self.reader.on_keyup(self._repl_key)
        self.reader.on_click(self._repl_select)
        self.reader.start()
        self.input.command_focus()
        self.started = True

    def query(self, eval=None):
        while True:
            message = self.reader.read()
            if not self.handle_message(message, eval):
                continue
            break

    def __hrepr__(self, H, hrepr):
        return H.div['repl-box'](
            H.bucheLog(address=self.relative_log_address),
            H.bucheInput(address='input'),
            style="height:100%;width:100%;display:flex;flex-direction:column",
            address=self.address
        )




import os
import sys
from .buche import *
from .repr import *
from .event import *
from .debug import *
from .repl import *


def _print_flush(x):
    print(x, flush=True)


def smart_breakpoint():
    if os.environ.get('BUCHE'):
        os.environ['PYTHONBREAKPOINT'] = 'buche.breakpoint'


H = hrepr.H
master = MasterBuche(hrepr, _print_flush)
buche = Buche(master, '/')
stdbuche = Buche(master, '/stdout')
reader = Reader(sys.stdin)
read = reader.read
repl = Repl(buche, reader)
breakpoint = BucheDb(repl).set_trace



"""
Setup for buche
"""

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='buche',
    version='0.2.0',

    description='Pretty-print Python objects to Buche logger.',
    long_description=long_description,

    url='https://github.com/breuleux/buche',

    author='Olivier Breuleux',
    author_email='breuleux@gmail.com',

    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='buche repr pprint html development',

    packages=find_packages(exclude=['contrib', 'doc', 'tests']),

    install_requires=['hrepr'],

    extras_require={
        'dev': [],
        'test': [],
    },

    # package_data={
    #     'buche': ['style/*.css'],
    # },

    python_requires='>=3.6',
)




import asyncio
import json
import re
from threading import Thread
from uuid import uuid4 as uuid
from hrepr import Tag
from .event import EventDispatcher
from .repr import id_registry


class Encoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Tag):
            return str(obj)
        else:
            super().default(obj)


encoder = Encoder()


class BucheMessage:
    def __init__(self, d):
        self.__dict__.update(d)

    def __hrepr__(self, H, hrepr):
        return hrepr.stdrepr_object('BucheMessage', self.__dict__.items())


class BucheSeq:
    def __init__(self, objects):
        self.objects = list(objects)

    @classmethod
    def __hrepr_resources__(cls, H):
        return H.style('''
        .multi-print { display: flex; align-items: center; }
        .multi-print > * { margin-right: 10px; }
        ''')

    def __hrepr__(self, H, hrepr):
        return H.div['multi-print'](*map(hrepr, self.objects))


class BucheDict:
    def __init__(self, keys):
        if hasattr(keys, 'items'):
            self.keys = keys.items()
        else:
            self.keys = keys

    def __hrepr__(self, H, hrepr):
        return hrepr.stdrepr_object(None, self.keys)


class MasterBuche:
    def __init__(self, hrepr, write):
        self.hrepr = hrepr
        self.write = write
        self.resources = set()

    def require(self, plugin):
        self.send(command='plugin', name=plugin)

    def send(self, d={}, **params):
        message = {**d, **params}
        self.write(encoder.encode(message))

    def generate(self, obj, hrepr_params={}):
        x = self.hrepr(obj, **hrepr_params)
        for res in self.hrepr.resources - self.resources:
            if res.name == 'buche-require':
                self.send(
                    command = 'plugin',
                    name = res.attributes['name']
                )
            else:
                self.send(
                    command = 'resource',
                    content = str(res)
                )
            self.resources.add(res)
        return x

    def show(self, obj, hrepr_params={}, **params):
        x = self.generate(obj, hrepr_params)
        self.send(format='html', content=str(x), **params)


class HTMLSender:
    def __init__(self, buche, current, open=False, repr=False):
        self.__buche = buche
        self.__current = current
        self.__open = open
        self.__repr = repr

    def __getattr__(self, attr):
        return HTMLSender(self.__buche,
                          getattr(self.__current, attr),
                          self.__open,
                          self.__repr)

    def __getitem__(self, item):
        return HTMLSender(self.__buche,
                          self.__current[item],
                          self.__open,
                          self.__repr)

    def __call__(self, *args, **kwargs):
        args = [self.__buche.master.generate(arg, kwargs)
                if self.__repr else arg
                for arg in args]
        address = None
        if isinstance(self.__current, Tag):
            if self.__repr:
                res = self.__current(*args)
            else:
                res = self.__current(*args, **kwargs)
            if self.__open and 'address' not in res.attributes:
                res = res(address=f'/{str(uuid())}')
            address = res.attributes.get('address', None)
        else:
            res = "".join(map(str, args))
        self.__buche.send(content=str(res))
        if self.__open:
            if address is not None:
                return self.__buche[address]
            else:
                raise Exception('Buche.open cannot be called directly.')


class Buche:
    def __init__(self, master, parent):
        self.master = master
        self.parent = parent

    def require(self, plugin, **opts):
        self.master.require(plugin, **opts)

    def send(self, parent=None, **params):
        if parent is None:
            cmd = params.get('command', 'log')
            if cmd in {'template', 'plugin', 'redirect', 'resource'}:
                self.master.send(**params)
                return
            parent = self.parent
        elif parent.startswith('/'):
            pass
        else:
            parent = self.join_path(parent)
        self.master.send(parent=parent, **params)

    @property
    def html(self):
        return HTMLSender(self, self.master.hrepr.H, False)

    @property
    def show(self):
        return HTMLSender(self, self.master.hrepr.H, False, True)

    @property
    def open(self):
        return HTMLSender(self, self.master.hrepr.H, True)

    def join_path(self, p):
        return f'{self.parent.rstrip("/")}/{p.strip("/")}'

    def __getattr__(self, attr):
        if attr.startswith('command_'):
            def cmd(**kwargs):
                self.send(
                    command=attr[8:],
                    **kwargs
                )
            return cmd
        else:
            return getattr(super(), attr)

    def __getitem__(self, item):
        if not item.startswith('/'):
            item = self.join_path(item)
        return Buche(self.master, item)

    def dict(self, **keys):
        self(BucheDict(keys))

    def __call__(self, *objs, **hrepr_params):
        if len(objs) == 1:
            o, = objs
        else:
            o = BucheSeq(objs)
        self.master.show(o, hrepr_params=hrepr_params, parent=self.parent)


class Reader(EventDispatcher):
    def __init__(self, source):
        super().__init__()
        self.ev_loop = asyncio.get_event_loop()
        self.source = source
        self.thread = Thread(target=self.loop)
        self.futures = []

    def read(self):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.read_async())

    def parse(self, line):
        d = json.loads(line)
        if 'argument' in d and 'obj' not in d:
            arg = d['argument']
            if arg:
                d['obj'] = id_registry.resolve(int(arg))
        message = BucheMessage(d)
        self.emit(d.get('eventType', 'UNDEFINED'), message)
        # Provide the message to all the coroutines waiting for one
        futs, self.futures = self.futures, []
        for fut in futs:
            self.ev_loop.call_soon_threadsafe(fut.set_result, message)
        return message

    def __iter__(self):
        for line in self.source:
            yield self.parse(line)

    def loop(self):
        for _ in self:
            pass

    def start(self):
        if not self.thread.is_alive():
            self.thread.start()

    def read_async(self):
        fut = asyncio.Future()
        self.futures.append(fut)
        return fut
